package org.openjdk.gcbench.runtime.writes;

import org.openjdk.jmh.annotations.*;

import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(1)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.NANOSECONDS)
@Threads(1)
@State(Scope.Thread)
public class WriteBarriersRef {

    Target target;
    Target source;

    @Setup
    public void setup() {
        target = new Target();
        source = new Target();
    }

    @Benchmark

    public void test() {
        doStore(target, source);
    }

    @CompilerControl(CompilerControl.Mode.DONT_INLINE)
    private void doStore(Target t, Target v) {
        t.field = v;
    }

    static class Target {
        Object field;
    }

    /*
       i7 4790K, 4.0 Ghz, Linux x86_64, JDK 9 (Shenandoah, 2016-09-05)

           Benchmark              Mode  Cnt  Score   Error  Units

           # Shenandoah
           WriteBarriersRef.test  avgt   25  3.222 ± 0.003  ns/op

           # G1
           WriteBarriersRef.test  avgt   25  2.341 ± 0.005  ns/op

           # Parallel
           WriteBarriersRef.test  avgt   25  2.176 ± 0.006  ns/op

        The difference is in a more complicated write barrier that may be improved a little with more
        implicit null checks and code quality tuneups.

        Shenandoah:

                               [Verified Entry Point]
          5.12%    4.76%         0x00007f51f0845dc0: mov    %eax,-0x14000(%rsp)
          3.00%    2.29%         0x00007f51f0845dc7: push   %rbp
          0.08%    0.05%         0x00007f51f0845dc8: sub    $0x10,%rsp
          4.90%    3.93%         0x00007f51f0845dcc: mov    -0x8(%rsi),%r10     ; <--- read barrier
          2.34%    0.63%         0x00007f51f0845dd0: mov    0xc(%r10),%r11d     ; get field $target
          0.05%    0.08%         0x00007f51f0845dd4: mov    0x10(%r10),%r10d    ; <-- get field $source
          4.68%    5.24%         0x00007f51f0845dd8: test   %r11d,%r11d         ; nullcheck $target <--- NOT IMPLICIT?
                                 0x00007f51f0845ddb: je     0x00007f51f0845e6c
          0.05%    0.03%         0x00007f51f0845de1: test   %r10d,%r10d         ; nullcheck $source <--- NOT IMPLICIT?
                          ╭      0x00007f51f0845de4: je     0x00007f51f0845e33
          2.56%    0.19%  │      0x00007f51f0845de6: shl    $0x3,%r10           ; unpack $source
          0.03%           │      0x00007f51f0845dea: mov    -0x8(%r10),%rbx     ; <--- HUH? Is this a null check trap? Unpacking is not needed then.
          6.58%    5.81%  │  ↗   0x00007f51f0845dee: lea    (%r12,%r11,8),%r10  ; unpack $target
                          │  │   0x00007f51f0845df2: mov    -0x8(%r10),%rbp     ; <--- read barrier
          2.00%    0.06%  │  │   0x00007f51f0845df6: cmpb   $0x0,0x3d8(%r15)    ; "evacuation in progress?"
                          │  │   0x00007f51f0845dfe: mov    -0x8(%r10),%rbp     ; <--- read barrier again (in sequence)
          6.46%    6.90%  │╭ │   0x00007f51f0845e02: je     0x00007f51f0845e0f  ; no evacuation? jump over the barrier <--- SHOULD MOVE TO SLOWPATH
                          ││ │   0x00007f51f0845e04: xchg   %rax,%rbp
                          ││ │   0x00007f51f0845e07: callq  Stub::shenandoah_wb ;   {runtime_call StubRoutines (2)}
                          ││ │   0x00007f51f0845e0c: xchg   %rax,%rbp
                          │↘ │   0x00007f51f0845e0f: movsbl 0x378(%r15),%r11d   ; SATB check and jump
          1.29%    1.33%  │  │   0x00007f51f0845e17: test   %r11d,%r11d
                          │ ╭│   0x00007f51f0845e1a: jne    0x00007f51f0845e37
                          │ ││↗  0x00007f51f0845e1c: mov    %rbx,%r10           ; packing $source (note, we unpacked before)
          6.61%    6.90%  │ │││  0x00007f51f0845e1f: shr    $0x3,%r10
                          │ │││  0x00007f51f0845e23: mov    %r10d,0xc(%rbp)     ; field store!
         18.55%   20.33%  │ │││  0x00007f51f0845e27: add    $0x10,%rsp
                          │ │││  0x00007f51f0845e2b: pop    %rbp
          2.93%    3.37%  │ │││  0x00007f51f0845e2c: test   %eax,0x11e691ce(%rip)
                          │ │││  0x00007f51f0845e32: retq


        Parallel:

                           [Verified Entry Point]
          3.65%    4.20%     0x00007f5169af3040: mov    %eax,-0x14000(%rsp)
          9.31%    9.50%     0x00007f5169af3047: push   %rbp
                             0x00007f5169af3048: sub    $0x10,%rsp
          3.17%    4.53%     0x00007f5169af304c: mov    0x10(%rsi),%r11d        ; get field $source
          8.67%    7.93%     0x00007f5169af3050: mov    0xc(%rsi),%r10d         ; get field $target
                          ╭  0x00007f5169af3054: mov    %r11d,0xc(%r12,%r10,8)  ; field store, implicit exception: dispatches to 0x00007f5169af307b
          8.29%    8.77%  │  0x00007f5169af3059: shl    $0x3,%r10               ; card mark update
                          │  0x00007f5169af305d: shr    $0x9,%r10
          6.58%    5.85%  │  0x00007f5169af3061: movabs $0x7f517632f000,%r11
                          │  0x00007f5169af306b: mov    %r12b,(%r11,%r10,1)
         11.76%   11.87%  │  0x00007f5169af306f: add    $0x10,%rsp              ; epilog
                          │  0x00007f5169af3073: pop    %rbp
          5.28%    5.76%  │  0x00007f5169af3074: test   %eax,0x145b3f86(%rip)
                          │  0x00007f5169af307a: retq
                          ↘  0x00007f5169af307b: mov    $0xfffffff6,%esi

     */

}
